from .api import (
    APIError,
    HTTPError,
    ValidationError,
    NotFoundError,
    UnauthorizedError,
    ForbiddenError,
    InternalServerError
)
from .client import InitializationError
from .page import PageError
from .parser import ParseException
